from django.db import migrations


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.RunSQL(
            "CREATE EXTENSION IF NOT EXISTS ltree;", "DROP EXTENSION ltree;"
        )
    ]



from django.db import models
from typing import TYPE_CHECKING
from django_ltree.paths import PathGenerator

from .querysets import TreeQuerySet

if TYPE_CHECKING:
    from django_ltree.models import TreeModel


class TreeManager(models.Manager):
    def get_queryset(self) -> TreeQuerySet["TreeModel"]:
        """Returns a queryset with the models ordered by `path`"""
        return TreeQuerySet(model=self.model, using=self._db).order_by("path")

    def roots(self) -> TreeQuerySet["TreeModel"]:
        """Returns the roots of a given model"""
        return self.filter().roots()

    def children(self, path: str) -> TreeQuerySet["TreeModel"]:
        """Returns the childrens of a given object"""
        return self.filter().children(path)

    def create_child(
        self, parent: "TreeModel" = None, label: str = None, **kwargs
    ) -> TreeQuerySet["TreeModel"]:
        """Creates a tree child with or without parent"""
        prefix = parent.path if parent else None

        """If a label is not provided, we generate a new one, else we use it as suffix"""
        if label is None:
            paths_in_use = parent.children() if parent else self.roots()
            path_generator = PathGenerator(
                prefix,
                skip=paths_in_use.values_list("path", flat=True),
            )
            path = next(path_generator)
        else:
            if prefix is None:
                path = label
            else:
                path = str(prefix) + "." + label

        kwargs["path"] = path
        return self.create(**kwargs)



from django.db.models import Transform
from django.db.models import fields
from .fields import PathField


__all__ = ("NLevel",)


@PathField.register_lookup
class NLevel(Transform):
    lookup_name = "depth"
    function = "nlevel"

    @property
    def output_field(self):
        return fields.IntegerField()



from django.db import models

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .models import TreeModel


class TreeQuerySet(models.QuerySet):
    def roots(self) -> models.QuerySet["TreeModel"]:
        return self.filter(path__depth=1)

    def children(self, path: str) -> models.QuerySet["TreeModel"]:
        return self.filter(path__descendants=path, path__depth=len(path) + 1)



from collections import UserList
from django import forms
from django.core.validators import RegexValidator
from django.db.models.fields import TextField
from django.forms.widgets import TextInput

from collections.abc import Iterable

path_label_validator = RegexValidator(
    r"^(?P<root>[a-zA-Z0-9_-]+)(?:\.[a-zA-Z0-9_-]+)*$",
    "A label is a sequence of alphanumeric characters and underscores separated by dots.",
    "invalid",
)


class PathValue(UserList):
    def __init__(self, value):
        if isinstance(value, str):
            split_by = "/" if "/" in value else "."
            value = value.strip().split(split_by) if value else []
        elif isinstance(value, int):
            value = [str(value)]
        elif isinstance(value, Iterable):
            value = [str(v) for v in value]
        else:
            raise ValueError("Invalid value: {!r} for path".format(value))

        super().__init__(initlist=value)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return ".".join(self)

    def __hash__(self):
        return hash(tuple(self))


class PathValueProxy:
    def __init__(self, field_name):
        self.field_name = field_name

    def __get__(self, instance, owner):
        if instance is None:
            return self

        value = instance.__dict__[self.field_name]

        if value is None:
            return value

        return PathValue(instance.__dict__[self.field_name])

    def __set__(self, instance, value):
        if instance is None:
            return self

        instance.__dict__[self.field_name] = value


class PathFormField(forms.CharField):
    default_validators = [path_label_validator]


class PathField(TextField):
    default_validators = [path_label_validator]

    def db_type(self, connection):
        return "ltree"

    def formfield(self, **kwargs):
        kwargs["form_class"] = PathFormField
        kwargs["widget"] = TextInput(attrs={"class": "vTextField"})
        return super().formfield(**kwargs)

    def contribute_to_class(self, cls, name, private_only=False):
        super().contribute_to_class(cls, name)
        setattr(cls, self.name, PathValueProxy(self.name))

    def from_db_value(self, value, expression, connection, *args):
        if value is None:
            return value
        return PathValue(value)

    def get_prep_value(self, value):
        if value is None:
            return value
        return str(PathValue(value))

    def to_python(self, value):
        if value is None:
            return value
        elif isinstance(value, PathValue):
            return value

        return PathValue(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        if value is None:
            return value
        elif isinstance(value, PathValue):
            return str(value)
        elif isinstance(value, (list, str)):
            return str(PathValue(value))

        raise ValueError(f"Unknown value type {type(value)}")
